/*
 * Moschip 7840 USB to serial adaptor driver
 *
 * Copyright (C) 2001-2002 Greg Kroah-Hartman (greg@kroah.com)
 * 
 * Support added for Moschip adaptor by ravikanth G.
 *
 * Original driver for 2.2.x by anonymous
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 * See Documentation/usb/usb-serial.txt for more information on using this driver
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/tty.h>
#include <linux/tty_driver.h>
#include <linux/tty_flip.h>
#include <linux/serial.h>
#include <linux/module.h>
#include <linux/spinlock.h>
#include <asm/uaccess.h>
#include <linux/usb.h>
#include "mos78xx.h"
#include "mos78xx_16C50.h"

#ifdef CONFIG_USB_SERIAL_DEBUG
static int debug = 1;
#else
static int debug;
#endif

#include "usb-serial.h"

/*
 * Version Information
 */
#define DRIVER_VERSION "v1.0.0.0"
#define DRIVER_DESC "MOSCHIP 7840/20 USB to serial adaptor driver"

static struct usb_device_id moschip_port_id_table[] = {
	{USB_DEVICE(USB_VENDOR_ID_MOSCHIP, MOSCHIP_DEVICE_ID_7840)},
	{USB_DEVICE(USB_VENDOR_ID_MOSCHIP, MOSCHIP_DEVICE_ID_7820)},
	{}                      /* terminating entry */
};

MODULE_DEVICE_TABLE (usb, moschip_port_id_table);

/* function prototypes for a MOSCHIP serial converter */
static int mos78xx_open (struct usb_serial_port *port, struct file *filp);
static void mos78xx_close (struct usb_serial_port *port, struct file *filp);
static void mos78xx_set_termios (struct usb_serial_port *port,struct termios *old);
static int mos78xx_ioctl (struct usb_serial_port *port, struct file *file,unsigned int cmd, unsigned long arg);
static int mos78xx_write (struct usb_serial_port *port, int from_user,const unsigned char *buf, int count);
static int  mos78xx_write_room(struct usb_serial_port *port);
static void mos78xx_throttle(struct usb_serial_port *port);
static void mos78xx_unthrottle(struct usb_serial_port *port);
static void mos78xx_break_ctl(struct usb_serial_port *port,int break_state);
static int mos78xx_startup (struct usb_serial *serial);
static void mos78xx_shutdown (struct usb_serial *serial);
static int mos78xx_chars_in_buffer (struct usb_serial_port *port);

static void mos78xx_read_int_callback (struct urb *urb);
static void mos78xx_read_bulk_callback (struct urb *urb);
static void mos78xx_write_bulk_callback (struct urb *urb);
static void mos78xx_control_callback (struct urb *urb);

int mos78xx_handle_newLsr(struct moschip_port *port,__u8 newLsr);
int mos78xx_handle_newMsr(struct moschip_port *port,__u8 newMsr);

static int get_serial_info(struct moschip_port *mos78xx_port, struct serial_struct * retinfo);
static int get_modem_info(struct moschip_port *mos78xx_port, unsigned int *value);
static int set_modem_info(struct moschip_port *mos78xx_port, unsigned int cmd, unsigned int *value);
static int get_lsr_info(struct moschip_port *mos78xx_port, unsigned int *value);
static int get_number_bytes_avail(struct moschip_port *mos78xx_port, unsigned int *value);

static void mos78xx_change_port_settings (struct moschip_port *mos78xx_port, struct termios *old_termios);
static void mos78xx_block_until_chase_response(struct moschip_port *mos78xx_port);
static void mos78xx_block_until_tx_empty (struct moschip_port *mos78xx_port);
static int mos78xx_set_reg_sync(struct usb_serial_port *port, __u16 reg, __u16 val);
static int mos78xx_get_reg_sync(struct usb_serial_port *port, __u16 reg, __u16 * val);
static int mos78xx_get_reg(struct moschip_port *mcs,__u16 Wval, __u16 reg, __u16 * val);

static int mos78xx_get_Uart_Reg(struct usb_serial_port *port, __u16 reg, __u16 * val);
static int mos78xx_set_Uart_Reg(struct usb_serial_port *port, __u16 reg, __u16
val);
static int mos78xx_send_cmd_write_baud_rate (struct moschip_port *mos78xx_port,
int baudRate);
static int mos78xx_calc_baud_rate_divisor (int baudRate, int *divisor,__u16 *clk_sel_val);
static int mos78xx_calc_num_ports(struct usb_serial *serial);
int RS485mode=0; //set to 1 for RS485 mode and 0 for RS232 mode

/* All of the device info needed for the PL2303 SIO serial converter */
static struct usb_serial_device_type mos78xx_device = {
	.owner =		THIS_MODULE,
	.name =			"MOS-7840/20",
	.id_table =		moschip_port_id_table,
	.num_interrupt_in =	1,
	.num_bulk_in =		4,
	.num_bulk_out =		4,
	.num_ports =		4,
	.open =			mos78xx_open,
	.close =		mos78xx_close,
	.write =		mos78xx_write,
	.write_room = 		mos78xx_write_room,
	.chars_in_buffer = 	mos78xx_chars_in_buffer,
	.throttle = 		mos78xx_throttle,
        .unthrottle = 		mos78xx_unthrottle,
	.ioctl =		mos78xx_ioctl,
	.break_ctl =		mos78xx_break_ctl,
	.set_termios =		mos78xx_set_termios,
	.read_bulk_callback =	mos78xx_read_bulk_callback,
	.read_int_callback =	mos78xx_read_int_callback,
	.write_bulk_callback =	mos78xx_write_bulk_callback,
	.startup =		mos78xx_startup,
	.shutdown =		mos78xx_shutdown,
};

struct mos78xx_private { 
	u8 line_control;
	u8 termios_initialized;
};

static int mos78xx_startup (struct usb_serial *serial)
{
	struct moschip_serial *mos78xx_serial;
	struct moschip_port *mos78xx_port;
	struct usb_device *dev;
	int i,status;
	__u16 Data;
	dev = serial->dev;
	mos78xx_serial = kmalloc(sizeof(struct moschip_serial), GFP_KERNEL);
	if(mos78xx_serial == NULL)
	{
		printk("%s - Out of memory", __FUNCTION__);
		return -ENOMEM;
	}
	memset (mos78xx_serial, 0, sizeof(struct moschip_serial));
	mos78xx_serial->serial = serial;
	serial->private = mos78xx_serial;
	mos78xx_serial->mos78xx_spectrum_2or4ports = mos78xx_calc_num_ports(serial);
	for(i=0; i< serial->num_ports; ++i)
	{
		mos78xx_port = kmalloc(sizeof(struct moschip_port), GFP_KERNEL);
		if(mos78xx_port == NULL)	
		{
			printk("%s Port out of memory",__FUNCTION__);
			serial->private = NULL;
			kfree(mos78xx_serial);
			return -ENOMEM;
		}
		memset(mos78xx_port, 0, sizeof(struct moschip_port));
		mos78xx_port->port = &serial->port[i];
		serial->port[i].private = mos78xx_port;
		mos78xx_port->port_num = serial->port[i].number + 1;
		if(mos78xx_port->port_num ==1)
		{

			mos78xx_port->ControlRegOffset =0x1;
			mos78xx_port->DcrRegOffset =0x4 ;
		}
		else if((mos78xx_port->port_num ==2)&&(mos78xx_serial->mos78xx_spectrum_2or4ports ==4))
		{
			mos78xx_port->SpRegOffset =0x8;
			mos78xx_port->ControlRegOffset =0x9;
			mos78xx_port->DcrRegOffset =0x16;
		}
		else if((mos78xx_port->port_num ==2)&&(mos78xx_serial->mos78xx_spectrum_2or4ports ==2))
		{
			mos78xx_port->SpRegOffset =0xa;
			mos78xx_port->ControlRegOffset =0xb;
			mos78xx_port->DcrRegOffset =0x19;
		}
		else if((mos78xx_port->port_num ==3)&&(mos78xx_serial->mos78xx_spectrum_2or4ports ==4))
		{
			mos78xx_port->SpRegOffset =0xa;
			mos78xx_port->ControlRegOffset =0xb;
			mos78xx_port->DcrRegOffset =0x19;
		}
		else if((mos78xx_port->port_num ==4)&&(mos78xx_serial->mos78xx_spectrum_2or4ports ==4))
		{
			mos78xx_port->SpRegOffset =0xc;
			mos78xx_port->ControlRegOffset =0xd;
			mos78xx_port->DcrRegOffset =0x1c ;
		}
		//enable rx_disable bit in control register
		status=mos78xx_get_reg_sync(&serial->port[i],mos78xx_port->ControlRegOffset,&Data);
		if(status<0) {
			DPRINTK("Reading ControlReg failed status-0x%x\n", status);
			break;
		}
		else 
			DPRINTK("ControlReg Reading success val is %x, status%d\n",Data,status);

		Data |= 0x08;//setting driver done bit
		Data |= 0x04;//sp1_bit to have cts change reflect in modem status reg
		//Data |= 0x20; //rx_disable bit

		status=0;
		status=mos78xx_set_reg_sync(&serial->port[i],mos78xx_port->ControlRegOffset,Data);
		if(status<0) {
			DPRINTK("Writing ControlReg failed(rx_disable) status-0x%x\n", status);
			break;
		}
		else 
			DPRINTK("ControlReg Writing success(rx_disable) status%d\n",status);

		//Write default values in DCR (i.e 0x01 in DCR0, 0x05 in DCR2 and 0x24 in DCR3
		Data = 0x01;
		status=0;
		status=mos78xx_set_reg_sync(&serial->port[i],(__u16)(mos78xx_port->DcrRegOffset+0),Data);
		if(status<0) {
			DPRINTK("Writing DCR0 failed status-0x%x\n", status);
			break;
		}
		else 
			DPRINTK("DCR0 Writing success status%d\n",status);

		Data = 0x05;
		status=0;
		status=mos78xx_set_reg_sync(&serial->port[i],(__u16)(mos78xx_port->DcrRegOffset+1),Data);
		if(status<0) {
			DPRINTK("Writing DCR1 failed status-0x%x\n", status);
			break;
		}
		else 
			DPRINTK("DCR1 Writing success status%d\n",status);

		Data = 0x24;
		status=0;
		status=mos78xx_set_reg_sync(&serial->port[i],(__u16)(mos78xx_port->DcrRegOffset+2),Data);
		if(status<0) {
			DPRINTK("Writing DCR2 failed status-0x%x\n", status);
			break;
		}
		else 
			DPRINTK("DCR2 Writing success status%d\n",status);

		// write values in clkstart0x0 and clkmulti 0x20
		Data = 0x0;
		status=0;
		status=mos78xx_set_reg_sync(&serial->port[i],CLK_START_VALUE_REGISTER,Data);
		if(status<0) {
			DPRINTK("Writing CLK_START_VALUE_REGISTER failed status-0x%x\n", status);
			break;
		}
		else 
			DPRINTK("CLK_START_VALUE_REGISTER Writing success status%d\n",status);


		Data = 0x20;
		status=0;
		status=mos78xx_set_reg_sync(&serial->port[i],CLK_MULTI_REGISTER,Data);
		if(status<0) {
			DPRINTK("Writing CLK_MULTI_REGISTER failed status-0x%x\n", status);
			break;
		}
		else 
			DPRINTK("CLK_MULTI_REGISTER Writing success status%d\n",status);

		mos78xx_port->control_urb = usb_alloc_urb(0);
		mos78xx_port->ctrl_buf = kmalloc(16,GFP_KERNEL);

	}

	//Zero Length flag enable
	Data = 0x0f;
	status=0;
	status=mos78xx_set_reg_sync(&serial->port[0],ZLP_REG5,Data);
	if(status<0) {
		DPRINTK("Writing ZLP_REG5 failed status-0x%x\n",status);
		return -1;
	}
	else 
		DPRINTK("ZLP_REG5 Writing success status%d\n",status);

	/* setting configuration feature to one */
	usb_control_msg (serial->dev, usb_sndctrlpipe(serial->dev, 0), (__u8)0x03, 0x00,0x01,0x00, 0x00, 0x00, 5*HZ);

	return 0;
}

static void mos78xx_shutdown(struct usb_serial *serial)
{
	int i;
	struct moschip_port *mos78xx_port;
	dbg("%s", __FUNCTION__);

	/* stop reads and writes on all ports */
	for (i=0; i < serial->num_ports; ++i) {
		mos78xx_port = serial->port[i].private;
		kfree(mos78xx_port->ctrl_buf);
		usb_unlink_urb (mos78xx_port->control_urb);
		usb_free_urb   (mos78xx_port->control_urb);
		kfree (mos78xx_port);
		serial->port[i].private = NULL;
	}
	kfree (serial->private);
	serial->private = NULL;
}


static void mos78xx_read_int_callback (struct urb *urb)
{
	int result;
	int length ;
	struct moschip_port   *mos78xx_port;
	struct moschip_serial *mos78xx_serial;
	struct usb_serial *serial;
	__u16 Data;
	unsigned char *data;
	__u8 sp[5],st;
	int i;
	__u16 wval;
	mos78xx_serial= (struct moschip_serial *)urb->context;
	if(!urb) 
	{
		DPRINTK("%s","Invalid Pointer !!!!:\n");
		return;
	}

	switch (urb->status)
	{
		case 0:
			/* success */
			break;
		case -ECONNRESET:
		case -ENOENT:
		case -ESHUTDOWN:
			/* this urb is terminated, clean up */
			dbg("%s - urb shutting down with status: %d", __FUNCTION__, urb->status);
			return;
		default:
			dbg("%s - nonzero urb status received: %d", __FUNCTION__, urb->status);
			goto exit;
	}
	length = urb->actual_length;
	data = urb->transfer_buffer;

	//mos78xx_serial= (struct moschip_serial *)urb->context;
	//serial  = mos78xx_get_usb_serial(port,__FUNCTION__);
	serial = mos78xx_serial->serial;

	/* Moschip get 5 bytes
	 * Byte 1 IIR Port 1 (port.number is 0)
	 * Byte 2 IIR Port 2 (port.number is 1)
	 * Byte 3 IIR Port 3 (port.number is 2)
	 * Byte 4 IIR Port 4 (port.number is 3)
	 * Byte 5 FIFO status for both */

	if(length && length>5)
	{
		DPRINTK("%s \n","Wrong data !!!");
		return;
	}

	/* MATRIX */
	if(mos78xx_serial->mos78xx_spectrum_2or4ports == 4)
	{
		sp[0]=(__u8)data[0];
		sp[1]=(__u8)data[1];
		sp[2]=(__u8)data[2];
		sp[3]=(__u8)data[3];
		st=(__u8)data[4];
	}
	else
	{
		sp[0]=(__u8)data[0];
		sp[1]=(__u8)data[2];
		//sp[2]=(__u8)data[2];
		//sp[3]=(__u8)data[3];
		st=(__u8)data[4];

	}
	for(i=0;i<serial->num_ports;i++)
	{
		mos78xx_port = serial->port[i].private;
		if((mos78xx_serial->mos78xx_spectrum_2or4ports == 2) && (i != 0))
			wval = (((__u16)serial->port[i].number - (__u16)(serial->minor))+2)<<8;
		else
			wval = (((__u16)serial->port[i].number - (__u16)(serial->minor))+1)<<8;
		if(mos78xx_port->open != FALSE)
		{
			//printk("%s wval is:(for 7840) %x\n",__FUNCTION__,wval);

			if(sp[i] & 0x01)
			{
				DPRINTK("SP%d No Interrupt !!!\n",i);
			}
			else
			{
				switch(sp[i] & 0x0f)
				{
					case SERIAL_IIR_RLS:
						DPRINTK("Serial Port %d: Receiver status error or ",i);
						DPRINTK("address bit detected in 9-bit mode\n");
						mos78xx_port->MsrLsr=1;
						mos78xx_get_reg(mos78xx_port,wval,LINE_STATUS_REGISTER,&Data);
						break;
					case SERIAL_IIR_MS:
						DPRINTK("Serial Port %d: Modem status change\n",i);
						mos78xx_port->MsrLsr=0;
						mos78xx_get_reg(mos78xx_port,wval, MODEM_STATUS_REGISTER, &Data);
						break;
				}
			}
		}
	}

exit:
	if( mos78xx_serial->status_polling_started == FALSE )
		return;

	result = usb_submit_urb (urb);
	if (result)
	{
		printk("%s - Error %d submitting interrupt urb\n", __FUNCTION__, result);
	}

	return;
}


static void mos78xx_read_bulk_callback (struct urb *urb)
{

	int                     status,i;
	unsigned char           *data ;
	struct usb_serial       *serial;
	struct usb_serial_port  *port;
	struct moschip_serial   *mos78xx_serial;
	struct moschip_port     *mos78xx_port;
	struct tty_struct *tty;
	if(!urb)
	{
		DPRINTK("%s","Invalid Pointer !!!!:\n"); 
		return;
	}

	if (urb->status)
	{
		DPRINTK("nonzero read bulk status received: %d",urb->status);
		//              if(urb->status==84)
		//ThreadState=1;
		return;
	}
	mos78xx_port= (struct moschip_port*)urb->context;
	if(!mos78xx_port)
	{
		DPRINTK("%s","NULL mos78xx_port pointer	\n");
		return ;
	}

	port = (struct usb_serial_port *)mos78xx_port->port;
	dbg("%s - port %d", __FUNCTION__, port->number);
//	serial = get_usb_serial (port, __FUNCTION__);
	serial = port->serial;
	if (!serial)
		return;
	data = urb->transfer_buffer;
	mos78xx_serial = serial->private;

//	debug=1;
//	usb_serial_debug_data (__FILE__, __FUNCTION__, urb->actual_length, urb->transfer_buffer);
//	debug=0;

	if (urb->actual_length)
	{

		tty = mos78xx_port->port->tty;
		if(urb->actual_length==1)
		{
			if(data[0]==0x13)
				mos78xx_port->xoff_cond=0;
			else if(data[0]==0x11)
			{
				wake_up(&mos78xx_port->xoff_wait);
				mos78xx_port->xoff_cond=1;
			}
		}
		if (tty)
		{
			for (i = 0; i < urb->actual_length; ++i)
			{
				/* if we insert more than TTY_FLIPBUF_SIZE characters, we drop them. */
				if(tty->flip.count >= TTY_FLIPBUF_SIZE)
				{
					tty_flip_buffer_push(tty);
				}
				/* this doesn't actually push the data through unless tty->low_latency is set */
				tty_insert_flip_char(tty, data[i], 0);
			}
			tty_flip_buffer_push(tty);
		}

		mos78xx_port->icount.rx += urb->actual_length;
	}
	if(!mos78xx_port->read_urb)
	{
		printk("%s","URB KILLED !!!\n");
		return;
	}
	if(mos78xx_port->read_urb->status!=-EINPROGRESS)        {
		mos78xx_port->read_urb->dev = serial->dev;

		status = usb_submit_urb(mos78xx_port->read_urb);

		if (status)
		{
			printk(" usb_submit_urb(read bulk) failed, status = %d", status);
		}
	}

}


static void mos78xx_write_bulk_callback (struct urb *urb)
{
	struct moschip_port *mos78xx_port ;
	struct tty_struct *tty;
	if(!urb)
	{
		DPRINTK("%s","Invalid Pointer !!!!:\n");
		return;
	}

	if (urb->status)
	{
		DPRINTK("nonzero write bulk status received:%d\n", urb->status);
		return;
	}

	mos78xx_port = (struct moschip_port *)urb->context;
	if(!mos78xx_port)
	{
		DPRINTK("%s","NULL mos78xx_port pointer	\n");
		return ;
	}
	dbg("error %s ", __FUNCTION__);
	tty = mos78xx_port->port->tty;
	if (tty) {
		/* let the tty driver wakeup if it has a special write_wakeup function */
		if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) && tty->ldisc.write_wakeup) {
			(tty->ldisc.write_wakeup)(tty);
		}
		/* tell the tty driver that something has changed */
		wake_up_interruptible(&tty->write_wait);
	}
	mos78xx_port->write_in_progress = FALSE;
}


static int mos78xx_write (struct usb_serial_port *port, int from_user,
		const unsigned char *buf, int count)
{

	int status;
	int i;
	int bytes_sent = 0;
	int transfer_size;
	struct moschip_port *mos78xx_port;
	struct usb_serial *serial;
	struct moschip_serial *mos78xx_serial;
	struct urb    *urb;
	const unsigned char *current_position = buf;
	//unsigned char * data1;
	dbg("%s - port %d", __FUNCTION__, port->number);
//	serial = get_usb_serial (port, __FUNCTION__);
	serial = port->serial;
	if (!serial)
		return;

	mos78xx_serial = serial->private;
	mos78xx_port = port->private;
	if (mos78xx_port == NULL)
	{
		printk("%s mos78xx_port returned NULL\n",__FUNCTION__);
		return;
	}
	DPRINTK("mo78xx_port->flow_ctrl_type:%d\n",mos78xx_port->flow_ctrl_type);
	if(mos78xx_port->delta_cts_cond==0 && mos78xx_port->flow_ctrl_type==2)
	{

		//interruptible_sleep_on(&mos78xx_port->delta_cts_wait);
		wait_event_interruptible(mos78xx_port->delta_cts_wait,(mos78xx_port->delta_cts_cond==1));
	}

	if(mos78xx_port->xoff_cond==0 && mos78xx_port->flow_ctrl_type==1)
	{

		//interruptible_sleep_on(&mos78xx_port->delta_cts_wait);
		wait_event_interruptible(mos78xx_port->xoff_wait,(mos78xx_port->xoff_cond==1));
	}
	/* try to find a free urb in the list */
	urb = NULL;
	for (i = 0; i < NUM_URBS; ++i)
	{
		if (mos78xx_port->write_urb_pool[i]->status != -EINPROGRESS)
		{
			urb = mos78xx_port->write_urb_pool[i];
			break;
		}
	}

	if (urb == NULL)
	{
		dbg("%s - no more free urbs", __FUNCTION__);
		goto exit;
	}
	if (urb->transfer_buffer == NULL)
	{
		urb->transfer_buffer = kmalloc (URB_TRANSFER_BUFFER_SIZE, GFP_KERNEL);
		if (urb->transfer_buffer == NULL)
		{
			err("%s no more kernel memory...", __FUNCTION__);
			goto exit;
		}
	}

	transfer_size = min (count, URB_TRANSFER_BUFFER_SIZE);
	if (from_user)
	{
		if (copy_from_user (urb->transfer_buffer, current_position, transfer_size))  
		{
			bytes_sent = -EFAULT;
			goto exit;
		}
	}
	else
	{
		memcpy (urb->transfer_buffer, current_position, transfer_size);

	}

	if((mos78xx_serial->mos78xx_spectrum_2or4ports==2)&&((__u16)port->number != 0))
	{
		FILL_BULK_URB(urb,
				mos78xx_serial->serial->dev,
				usb_sndbulkpipe(mos78xx_serial->serial->dev,
					(port->bulk_out_endpointAddress)+2),
				urb->transfer_buffer,
				transfer_size,
				mos78xx_write_bulk_callback,
				mos78xx_port);
	}
	else
	{
		FILL_BULK_URB(urb,
				mos78xx_serial->serial->dev,
				usb_sndbulkpipe(mos78xx_serial->serial->dev,
					port->bulk_out_endpointAddress),
				urb->transfer_buffer,
				transfer_size,
				mos78xx_write_bulk_callback,
				mos78xx_port);
	}
//	data1=urb->transfer_buffer;
//	for(i=0;i < urb->actual_length;i++)
//	      printk("Data is %c\n ",data1[i]);

	/* send it down the pipe */
	status = usb_submit_urb(urb);
	if (status)
	{
		err("%s - write bulk failed with status= %d", __FUNCTION__, status);
		bytes_sent = status;
		goto exit;
	}
	bytes_sent = transfer_size;
	mos78xx_port->icount.tx += transfer_size;
exit:
	return bytes_sent;
}


static int  mos78xx_write_room(struct usb_serial_port *port)
{
	int i;
	int room = 0;
	struct moschip_port *mos78xx_port;

	mos78xx_port = port->private;
	if (mos78xx_port == NULL)
	{
		DPRINTK("%s \n","mos78xx_break:leaving ...........");
		return -1;
	}



	for (i = 0; i < NUM_URBS; ++i)
	{
		if (mos78xx_port->write_urb_pool[i]->status != -EINPROGRESS)
		{
			room += URB_TRANSFER_BUFFER_SIZE;
		}
	}

	dbg("%s - returns %d", __FUNCTION__, room);
	return (room);
}
     
static void mos78xx_throttle(struct usb_serial_port *port)
{
	struct moschip_port *mos78xx_port;
	struct tty_struct *tty;
	int status;
	mos78xx_port = port->private;
	if(mos78xx_port == NULL)
		return;
	if (!mos78xx_port->open)
	{
		DPRINTK("%s\n","port not opened");
		return;
	}
	tty = port->tty;
	if (!tty) {
		dbg ("%s - no tty available", __FUNCTION__);
		return;
	}

	/* if we are implementing XON/XOFF, send the stop character */
	if (I_IXOFF(tty)) {
		unsigned char stop_char = STOP_CHAR(tty);
		status = mos78xx_write (port, 0, &stop_char, 1);
		if (status <= 0) {
			return;
		}
	}
	/* if we are implementing RTS/CTS, toggle that line */
	if (tty->termios->c_cflag & CRTSCTS)
	{
		mos78xx_port->shadowMCR &= ~MCR_RTS;
		status=0;
		status=mos78xx_set_Uart_Reg(port,MODEM_CONTROL_REGISTER,mos78xx_port->shadowMCR);
		if (status < 0)
		{
			return;
		}
	}

	return;
}


static void mos78xx_unthrottle(struct usb_serial_port *port)
{
	struct moschip_port *mos78xx_port;
	struct tty_struct *tty;
	int status;
	mos78xx_port = port->private;
	if(mos78xx_port == NULL)
		return;
	if (!mos78xx_port->open)
	{
		DPRINTK("%s\n","port not opened");
		return;
	}
	tty = port->tty;
	if (!tty) {
		dbg ("%s - no tty available", __FUNCTION__);
		return;
	}
	/* if we are implementing XON/XOFF, send the start character */
	if (I_IXOFF(tty)) {
		unsigned char start_char = START_CHAR(tty);
		status = mos78xx_write (port, 0, &start_char, 1);
		if (status <= 0) {
			return;
		}
	}
	/* if we are implementing RTS/CTS, toggle that line */
	if (tty->termios->c_cflag & CRTSCTS) {
		mos78xx_port->shadowMCR |= MCR_RTS;
		status=0;
		status=mos78xx_set_Uart_Reg(port,MODEM_CONTROL_REGISTER,mos78xx_port->shadowMCR);
		if (status != 0) {
			return;
		}
	}
	return;
}

static int get_serial_info(struct moschip_port *mos78xx_port, struct serial_struct * retinfo)
{
	struct serial_struct tmp;
	if (mos78xx_port == NULL)
		return -1;
	if (!retinfo)
		return -EFAULT;

	memset(&tmp, 0, sizeof(tmp));

	tmp.type                = PORT_16550A;
	tmp.line                = mos78xx_port->port->serial->minor;
	tmp.port                = mos78xx_port->port->number;
	tmp.irq                 = 0;
	tmp.flags               = ASYNC_SKIP_TEST | ASYNC_AUTO_IRQ;
	tmp.xmit_fifo_size      = NUM_URBS * URB_TRANSFER_BUFFER_SIZE;
	tmp.baud_base           = 921600;//9600; //baud_base is taken based upon baud rate setting.
	//tmp.baud_base           = 9600; //baud_base is taken based upon baud rate setting.
	tmp.close_delay         = 5*HZ;
	tmp.closing_wait        = 30*HZ;
	if (copy_to_user(retinfo, &tmp, sizeof(*retinfo)))
		return -EFAULT;
	return 0;
}

static int get_modem_info(struct moschip_port *mos78xx_port, unsigned int *value)
{
	unsigned int result = 0;
	__u16 msr;
	unsigned int mcr = mos78xx_port->shadowMCR;
	int status=0;
	status=mos78xx_get_Uart_Reg(mos78xx_port->port,MODEM_STATUS_REGISTER,&msr);
	result = ((mcr & MCR_DTR)       ? TIOCM_DTR: 0)   /* 0x002 */
		| ((mcr & MCR_RTS)    ? TIOCM_RTS: 0)   /* 0x004 */
		| ((msr & MOS7840_MSR_CTS)    ? TIOCM_CTS: 0)   /* 0x020 */
		| ((msr & MOS7840_MSR_CD)     ? TIOCM_CAR: 0)   /* 0x040 */
		| ((msr & MOS7840_MSR_RI)     ? TIOCM_RI:  0)   /* 0x080 */
		| ((msr & MOS7840_MSR_DSR)    ? TIOCM_DSR: 0);  /* 0x100 */
	dbg("%s -- %x", __FUNCTION__, result);
	if (copy_to_user(value, &result, sizeof(int)))
		return -EFAULT;
	return 0;
}


static int set_modem_info(struct moschip_port *mos78xx_port, unsigned int cmd, unsigned int *value)
{
	unsigned int mcr ;
	unsigned int arg;
	__u16 Data;
	int status;
	struct usb_serial_port *port;

	if (mos78xx_port == NULL)
		return -1;
	port = mos78xx_port->port;
	mcr = mos78xx_port->shadowMCR;
	if (copy_from_user(&arg, value, sizeof(int)))
		return -EFAULT;
	switch (cmd) {
		case TIOCMBIS:
			if (arg & TIOCM_RTS)
				mcr |= MCR_RTS;
			if (arg & TIOCM_DTR)
				mcr |= MCR_RTS;
			if (arg & TIOCM_LOOP)
				mcr |= MCR_LOOPBACK;
			break;

		case TIOCMBIC:
			if (arg & TIOCM_RTS)
				mcr &= ~MCR_RTS;
			if (arg & TIOCM_DTR)
				mcr &= ~MCR_RTS;
			if (arg & TIOCM_LOOP)
				mcr &= ~MCR_LOOPBACK;
			break;
		case TIOCMSET:
			/* turn off the RTS and DTR and LOOPBACK
			 * and then only turn on what was asked to */
			mcr &=  ~(MCR_RTS | MCR_DTR | MCR_LOOPBACK);
			mcr |= ((arg & TIOCM_RTS) ? MCR_RTS : 0);
			mcr |= ((arg & TIOCM_DTR) ? MCR_DTR : 0);
			mcr |= ((arg & TIOCM_LOOP) ? MCR_LOOPBACK : 0);
			break;
	}

	mos78xx_port->shadowMCR = mcr;
	Data = mos78xx_port->shadowMCR;
	status=0;
	status = mos78xx_set_Uart_Reg(port,MODEM_CONTROL_REGISTER,Data);
	if(status <0)
	{
		DPRINTK("setting MODEM_CONTROL_REGISTER Failed\n");
		return -1;
	}

	return 0;
}	

static int get_lsr_info(struct moschip_port *mos78xx_port, unsigned int *value)
{
	int count;
	unsigned int result = 0;
	count = mos78xx_chars_in_buffer(mos78xx_port->port);
	if(count == 0)
	{
		dbg("%s -- Empty", __FUNCTION__);
		result = TIOCSER_TEMT;
	}

	if (copy_to_user(value, &result, sizeof(int)))
		return -EFAULT;
	return 0;

}

static int get_number_bytes_avail(struct moschip_port *mos78xx_port, unsigned int *value)
{
	unsigned int result = 0;
	struct tty_struct *tty = mos78xx_port->port->tty;
	if (!tty)
		return -ENOIOCTLCMD;
	result = tty->read_cnt;
	dbg("%s(%d) = %d", __FUNCTION__,  mos78xx_port->port->number, result);
	if (copy_to_user(value, &result, sizeof(int)))
		return -EFAULT;
	return -ENOIOCTLCMD;

}


static int mos78xx_ioctl (struct usb_serial_port *port, struct file *file,
		unsigned int cmd, unsigned long arg)
{

	struct moschip_port *mos78xx_port;
	struct async_icount cnow;
	struct async_icount cprev;
	struct serial_icounter_struct icount;
	struct tty_struct *tty;
	int mosret=0;

	mos78xx_port = port->private;
	tty = mos78xx_port->port->tty;
	if (mos78xx_port == NULL)
		return -1;

	dbg("%s - port %d, cmd = 0x%x", __FUNCTION__, port->number, cmd);

	switch (cmd)
	{
		/* return number of bytes available */
		case TIOCINQ:
			dbg("%s (%d) TIOCINQ", __FUNCTION__,  port->number);
			return get_number_bytes_avail(mos78xx_port, (unsigned int *) arg);
			break;

			//	case TIOCOUTQ:
			//		dbg("%s (%d) TIOCOUTQ", __FUNCTION__,  port->number);
			//		return put_user(tty->driver->chars_in_buffer ?
			//				tty->driver->chars_in_buffer(tty) : 0,
			//				(int __user *) arg);
			//		break;

		case TIOCSERGETLSR:
			dbg("%s (%d) TIOCSERGETLSR", __FUNCTION__,  port->number);
			return get_lsr_info(mos78xx_port, (unsigned int *) arg);
		case TIOCMBIS:
		case TIOCMBIC:
		case TIOCMSET:
			dbg("%s (%d) TIOCMSET/TIOCMBIC/TIOCMSET", __FUNCTION__,
					port->number);
			mosret=set_modem_info(mos78xx_port, cmd, (unsigned int *) arg);
			return mosret;
		case TIOCMGET:
			dbg("%s (%d) TIOCMGET", __FUNCTION__,  port->number);
			return get_modem_info(mos78xx_port, (unsigned int *) arg);
		case TIOCGSERIAL:
			dbg("%s (%d) TIOCGSERIAL", __FUNCTION__,  port->number);                        return get_serial_info(mos78xx_port, (struct serial_struct *) arg);
		case TIOCSSERIAL:
			dbg("%s (%d) TIOCSSERIAL", __FUNCTION__,  port->number);                        return 0;
			break;
		case TIOCMIWAIT:
			dbg("%s (%d) TIOCMIWAIT", __FUNCTION__,  port->number);
			cprev = mos78xx_port->icount;
			while (1) {
				interruptible_sleep_on(&mos78xx_port->delta_msr_wait);
				/* see if a signal did it */
				if (signal_pending(current))
					return -ERESTARTSYS;
				cnow = mos78xx_port->icount;
				if (cnow.rng == cprev.rng && cnow.dsr == cprev.dsr &&
						cnow.dcd == cprev.dcd && cnow.cts == cprev.cts)
					return -EIO; /* no change => error */
				if (((arg & TIOCM_RNG) && (cnow.rng != cprev.rng)) ||
						((arg & TIOCM_DSR) && (cnow.dsr != cprev.dsr)) ||
						((arg & TIOCM_CD)  && (cnow.dcd != cprev.dcd)) ||
						((arg & TIOCM_CTS) && (cnow.cts != cprev.cts)) ) {
					return 0;
				}
				cprev = cnow;
			}
			break;
		case TIOCGICOUNT:
			cnow = mos78xx_port->icount;
			icount.cts = cnow.cts;
			icount.dsr = cnow.dsr;
			icount.rng = cnow.rng;
			icount.dcd = cnow.dcd;
			icount.rx = cnow.rx;
			icount.tx = cnow.tx;
			icount.frame = cnow.frame;
			icount.overrun = cnow.overrun;
			icount.parity = cnow.parity;
			icount.brk = cnow.brk;
			icount.buf_overrun = cnow.buf_overrun;
			dbg("%s (%d) TIOCGICOUNT RX=%d, TX=%d", __FUNCTION__,  port->number, icount.rx, icount.tx );
			if (copy_to_user((void *)arg, &icount, sizeof(icount)))
				return -EFAULT;
			return 0;
		case TIOCEXBAUD:
			return 0;
		default:
			break;
	}
	return -ENOIOCTLCMD;
}

static void mos78xx_change_port_settings (struct moschip_port *mos78xx_port, struct termios *old_termios)
{
	struct tty_struct *tty;
	int baud;
	unsigned cflag;
	unsigned iflag;
	__u8 mask = 0xff;
	__u8 lData;
	__u8 lParity;
	__u8 lStop;
	int status;
	__u16 Data;
	struct usb_serial_port *port;
	struct usb_serial *serial;
	if (mos78xx_port == NULL)
	{
		printk("%s mos78xx_port returned NULL\n",__FUNCTION__);
		return;
	}
	dbg("%s - port %d", __FUNCTION__, port->number);
	port = (struct usb_serial_port *)mos78xx_port->port;
	serial = port->serial;
	if (!serial)
		return;
	tty = mos78xx_port->port->tty;
	if ((!tty) || (!tty->termios))
	{
		dbg("%s - no tty structures", __FUNCTION__);
		return;
	}
	lData = LCR_BITS_8;
	lStop = LCR_STOP_1;
	lParity = LCR_PAR_NONE;

	cflag = tty->termios->c_cflag;
	iflag = tty->termios->c_iflag;

	switch (cflag & CSIZE)
	{
		case CS5:       lData = LCR_BITS_5;
				mask = 0x1f;
				break;

		case CS6:       lData = LCR_BITS_6;
				mask = 0x3f;
				break;

		case CS7:       lData = LCR_BITS_7;
				mask = 0x7f;
				break;
		default:
		case CS8:       lData = LCR_BITS_8;
				break;
	}
	/* Change the Parity bit */
	if (cflag & PARENB)
	{
		if (cflag & PARODD)
		{
			lParity = LCR_PAR_ODD;
			dbg("%s - parity = odd", __FUNCTION__);
		}
		else
		{
			lParity = LCR_PAR_EVEN;
			dbg("%s - parity = even", __FUNCTION__);
		}

	}
	else
	{
		dbg("%s - parity = none", __FUNCTION__);
	}
	if(cflag & CMSPAR)
	{
		lParity = lParity | 0x20;
	}

	/* Change the Stop bit */
	if (cflag & CSTOPB)
	{
		lStop = LCR_STOP_2;
		dbg("%s - stop bits = 2", __FUNCTION__);
	}
	else
	{
		lStop = LCR_STOP_1;
		dbg("%s - stop bits = 1", __FUNCTION__);
	}
	/* Update the LCR with the correct value */
	mos78xx_port->shadowLCR &= ~(LCR_BITS_MASK | LCR_STOP_MASK | LCR_PAR_MASK);
	mos78xx_port->shadowLCR |= (lData | lParity | lStop);

	mos78xx_port->validDataMask = mask;
	DPRINTK("mos78xx_change_port_settings mos78xx_port->shadowLCR is %x\n",mos78xx_port->shadowLCR);
	/* Disable Interrupts */
	Data = 0x00;
	mos78xx_set_Uart_Reg(port,INTERRUPT_ENABLE_REGISTER,Data);


	Data = 0x00;
	mos78xx_set_Uart_Reg(port,FIFO_CONTROL_REGISTER,Data);

	Data = 0xcf;
	mos78xx_set_Uart_Reg(port,FIFO_CONTROL_REGISTER,Data);

	/* Send the updated LCR value to the mos78xx */
	Data = mos78xx_port->shadowLCR;

	mos78xx_set_Uart_Reg(port,LINE_CONTROL_REGISTER,Data);
	Data = 0x00b;
	mos78xx_port->shadowMCR = Data;
	mos78xx_set_Uart_Reg(port,MODEM_CONTROL_REGISTER,Data);
	Data = 0x00b;
	mos78xx_set_Uart_Reg(port,MODEM_CONTROL_REGISTER,Data);

	/* set up the MCR register and send it to the mos78xx */

	mos78xx_port->shadowMCR = MCR_MASTER_IE;
	if (cflag & CBAUD)
	{
		mos78xx_port->shadowMCR |= (MCR_DTR | MCR_RTS);
	}


	if (cflag & CRTSCTS)
	{
		mos78xx_port->shadowMCR |= (MCR_XON_ANY);


	}
	else
	{
		mos78xx_port->shadowMCR &= ~(MCR_XON_ANY);
	}
	Data = mos78xx_port->shadowMCR;
	mos78xx_set_Uart_Reg(port,MODEM_CONTROL_REGISTER,Data);
	/* Determine divisor based on baud rate */
	baud = tty_get_baud_rate(tty);
	if (!baud)
	{
		/* pick a default, any default... */
		DPRINTK("%s\n","Picked default baud...");
		baud = 9600;
	}
	dbg("%s - baud rate = %d", __FUNCTION__, baud);
	status = mos78xx_send_cmd_write_baud_rate (mos78xx_port, baud);

	/* Enable Interrupts */
	Data = 0x0c;
	mos78xx_set_Uart_Reg(port,INTERRUPT_ENABLE_REGISTER,Data);

	if(mos78xx_port->read_urb->status!=-EINPROGRESS)
	{
		mos78xx_port->read_urb->dev = serial->dev;
		status = usb_submit_urb(mos78xx_port->read_urb);
		if (status)
		{
			DPRINTK(" usb_submit_urb(read bulk) failed, status = %d", status);
		}
	}
	return;
}

static void mos78xx_set_termios (struct usb_serial_port *port, struct termios *old_termios)
{
	struct usb_serial *serial;
	struct moschip_port *mos78xx_port;
	struct tty_struct *tty = port->tty;
	int status;
	unsigned int cflag;
	dbg("%s - port %d", __FUNCTION__, port->number);
//	serial = get_usb_serial (port, __FUNCTION__);
	serial = port->serial;
	if (!serial)
		return;

	mos78xx_port = port->private;
	if (mos78xx_port == NULL)
	{
		printk("%s mos78xx_port returned NULL\n",__FUNCTION__);
		return;
	}

	if (!port->tty || !port->tty->termios)
	{
		dbg ("%s - no tty or termios", __FUNCTION__);
		return;
	}
	if (!mos78xx_port->open)
	{
		dbg("%s - port not opened", __FUNCTION__);
		return;
	}
	mos78xx_block_until_tx_empty(mos78xx_port);	
	cflag = tty->termios->c_cflag;
	DPRINTK("newterminal cflag=%u and iflag=%u\n",cflag,tty->termios->c_iflag);
	DPRINTK("oldterminal cflag=%u and iflag=%u\n",old_termios->c_cflag,old_termios->c_iflag);
	if (!cflag)
	{
		DPRINTK("%s %s\n",__FUNCTION__,"cflag is NULL");
		return;
	}

	/* check that they really want us to change something */
	if (old_termios)
	{
		if ((cflag == old_termios->c_cflag) &&
				(RELEVANT_IFLAG(tty->termios->c_iflag) == RELEVANT_IFLAG(old_termios->c_iflag)))
		{
			DPRINTK("%s\n","Nothing to change");
			return;
		}
	}
	dbg("%s - clfag %08x iflag %08x", __FUNCTION__,
			tty->termios->c_cflag,
			RELEVANT_IFLAG(tty->termios->c_iflag));

	if (old_termios)
	{
		dbg("%s - old clfag %08x old iflag %08x", __FUNCTION__,
				old_termios->c_cflag,
				RELEVANT_IFLAG(old_termios->c_iflag));
	}

	dbg("%s - port %d", __FUNCTION__, port->number);

	/* change the port settings to the new ones specified */
	mos78xx_change_port_settings (mos78xx_port, old_termios);
	if(!mos78xx_port->read_urb)
	{
		printk("%s","URB KILLED !!!!!\n");
		return;
	}

	if(mos78xx_port->read_urb->status!=-EINPROGRESS)
	{
		mos78xx_port->read_urb->dev = serial->dev;
		status = usb_submit_urb(mos78xx_port->read_urb);
		if (status)
		{
			printk(" usb_submit_urb(read bulk) failed, status = %d", status);
		}
	}
	return;
}

static void mos78xx_block_until_chase_response(struct moschip_port *mos78xx_port)
{
	int timeout = 1*HZ;
	int wait = 10;
	int count ;


	while (1)
	{
		count = mos78xx_chars_in_buffer(mos78xx_port->port);
		/* Check for Buffer status */
		if(count<=0)
		{
			mos78xx_port->chaseResponsePending = FALSE;
			return;
		}
		/* Block the thread for a while */
		interruptible_sleep_on_timeout (&mos78xx_port->wait_chase, timeout);
		/* No activity.. count down section */
		wait--;
		if (wait == 0)
		{
			dbg("%s - TIMEOUT", __FUNCTION__);
			return;
		}
		else
		{
			/* Reset timout value back to seconds */
			wait = 10;
		}
	}

}


static void mos78xx_break_ctl(struct usb_serial_port *port,int break_state)
{
	unsigned char data;
	struct usb_serial *serial;
	struct moschip_serial *mos78xx_serial;
	struct moschip_port *mos78xx_port;

	dbg("%s - port %d", __FUNCTION__, port->number);
//	serial = get_usb_serial (port, __FUNCTION__);
	serial = port->serial;
	if (!serial)
		return;

	mos78xx_port = port->private;
	if (mos78xx_port == NULL)
	{
		printk("%s mos78xx_port returned NULL\n",__FUNCTION__);
		return;
	}	
	mos78xx_serial = serial->private;
	if(mos78xx_serial == NULL)
	{
		printk("%s mos78xx_serial returned NULL\n",__FUNCTION__);
		return;
	}

	mos78xx_port->chaseResponsePending = TRUE;
	mos78xx_block_until_chase_response(mos78xx_port);
	if(break_state == -1)
	{
		data = mos78xx_port->shadowLCR | LCR_SET_BREAK;
	}
	else
	{
		data = mos78xx_port->shadowLCR & ~LCR_SET_BREAK;
	}

	mos78xx_port->shadowLCR  = data;
	DPRINTK("mcs7840_break mos78xx_port->shadowLCR is %x\n",mos78xx_port->shadowLCR);
	mos78xx_set_Uart_Reg(port,LINE_CONTROL_REGISTER,mos78xx_port->shadowLCR);
	return;
}

static int mos78xx_open (struct usb_serial_port *port, struct file * filp)
{
	struct usb_serial *serial;
	struct moschip_serial *mos78xx_serial;
	struct moschip_port *mos78xx_port;
	struct urb *urb;
	__u16 Data;
	int status,j;
	int response = 0;
	struct termios tmp_termios;
	serial = port->serial;
	if (!serial)
	{
		DPRINTK("serial is null is  null\n");
		return;
	}
	mos78xx_port = port->private;
	if (mos78xx_port == NULL)
	{
		printk("%s mos78xx_port returned NULL\n",__FUNCTION__);
		return -ENODEV;
	}	
	mos78xx_serial = serial->private;
	if(mos78xx_serial == NULL)
	{
		printk("%s mos78xx_serial returned NULL\n",__FUNCTION__);
		return -ENODEV;
	}
	mos78xx_serial->NoOfOpenPorts++;
	usb_clear_halt(serial->dev, port->write_urb->pipe);
	usb_clear_halt(serial->dev, port->read_urb->pipe);
	for (j = 0; j < NUM_URBS; ++j)
	{
		DPRINTK("Allocate memory for urb:%d\n",j);
		urb = usb_alloc_urb(0);
		if (urb == NULL)
		{
			printk("No more urbs???");
			continue;
		}
		mos78xx_port->write_urb_pool[j] = urb;
		urb->transfer_buffer = kmalloc (URB_TRANSFER_BUFFER_SIZE, GFP_KERNEL);
		memset(urb->transfer_buffer,0,URB_TRANSFER_BUFFER_SIZE);
		if (!urb->transfer_buffer)
		{
			printk("%s-out of memory for urb buffers.", __FUNCTION__);
			continue;
		}
	}
	status = 0x0;
	Data = 0x0;
	status = mos78xx_get_reg_sync(port,mos78xx_port->SpRegOffset,&Data);
	if(status < 0){
		DPRINTK("Reading Spreg failed\n");
		return -1;
	}
	Data |= 0x80;
	status = mos78xx_set_reg_sync(port,mos78xx_port->SpRegOffset,Data);
	if(status < 0){
		DPRINTK("writing Spreg failed\n");
		return -1;
	}
	Data &= ~0x80;
	status = mos78xx_set_reg_sync(port,mos78xx_port->SpRegOffset,Data);
	if(status < 0){
		DPRINTK("writing Spreg failed\n");
		return -1;
	}
	if(RS485mode==1)
		Data = 0xC0;
	else
		Data = 0x00;

	status=0;
	status=mos78xx_set_Uart_Reg(port,SCRATCH_PAD_REGISTER,Data);
	if(status<0) {
		DPRINTK("Writing SCRATCH_PAD_REGISTER failed status-0x%x\n", status);
		return -1;
	}
	else 
		DPRINTK("SCRATCH_PAD_REGISTER Writing success status%d\n",status);

	status=0;
	Data=0x0;
	status=mos78xx_get_reg_sync(port,mos78xx_port->ControlRegOffset,&Data);
	if(status<0){
		DPRINTK("Reading Controlreg failed\n");
		return -1;
	}
	Data |= 0x08;//Driver done bit
	Data |= 0x20;//rx_disable
	status=0;
	status = mos78xx_set_reg_sync(port,mos78xx_port->ControlRegOffset,Data);        if(status<0){
		DPRINTK("writing Controlreg failed\n");
		return -1;
	}
	//do register settings here
	// Set all regs to the device default values.
	////////////////////////////////////
	// First Disable all interrupts.
	////////////////////////////////////
	Data = 0x00;
	status=0;
	status = mos78xx_set_Uart_Reg(port,INTERRUPT_ENABLE_REGISTER,Data);
	if(status<0){
		DPRINTK("disableing interrupts failed\n");
		return -1;
	}
	// Set FIFO_CONTROL_REGISTER to the default value
	Data = 0x00;
	status=0;
	status = mos78xx_set_Uart_Reg(port,FIFO_CONTROL_REGISTER,Data);
	if(status<0){
		DPRINTK("Writing FIFO_CONTROL_REGISTER  failed\n");
		return -1;
	}
	Data = 0xcf;  //chk
	status=0;
	status = mos78xx_set_Uart_Reg(port,FIFO_CONTROL_REGISTER,Data);
	if(status<0){
		DPRINTK("Writing FIFO_CONTROL_REGISTER  failed\n");
		return -1;
	}
	Data = 0x03; //LCR_BITS_8
	status=0;
	status = mos78xx_set_Uart_Reg(port,LINE_CONTROL_REGISTER,Data);
	mos78xx_port->shadowLCR=Data;

	Data = 0x0b; // MCR_DTR|MCR_RTS|MCR_MASTER_IE
	status=0;
	status = mos78xx_set_Uart_Reg(port,MODEM_CONTROL_REGISTER,Data);
	mos78xx_port->shadowMCR=Data;

	//clearing Bulkin and Bulkout Fifo
	Data = 0x0;
	status = 0;
	status = mos78xx_get_reg_sync(port,mos78xx_port->SpRegOffset,&Data);

	Data = Data | 0x0c;
	status = 0;
	status = mos78xx_set_reg_sync(port,mos78xx_port->SpRegOffset,Data);

	Data = Data & ~0x0c;
	status = 0;
	status = mos78xx_set_reg_sync(port,mos78xx_port->SpRegOffset,Data);
	//Finally enable all interrupts
	Data = 0x0;
	Data = 0x0c;
	status = 0;
	status = mos78xx_set_Uart_Reg(port,INTERRUPT_ENABLE_REGISTER,Data);

	//clearing rx_disable
	Data = 0x0;
	status = 0;
	status = mos78xx_get_reg_sync(port,mos78xx_port->ControlRegOffset,&Data);

	Data = Data & ~0x20;
	status = 0;
	status = mos78xx_set_reg_sync(port,mos78xx_port->ControlRegOffset,Data);                                                                                
	// rx_negate
	Data = 0x0;
	status = 0;
	status = mos78xx_get_reg_sync(port,mos78xx_port->ControlRegOffset,&Data);
	Data = Data |0x10;
	status = 0;
	status = mos78xx_set_reg_sync(port,mos78xx_port->ControlRegOffset,Data);

	if(port->tty)
		port->tty->low_latency=1;
#ifdef interrupt_test
	if(mos78xx_serial->NoOfOpenPorts==1)
	{
		// start the status polling here
		mos78xx_serial->status_polling_started = TRUE;
		/* If not yet set, Set here */
		mos78xx_serial->interrupt_in_buffer = serial->port[0].interrupt_in_buffer;
		mos78xx_serial->interrupt_in_endpoint = serial->port[0].interrupt_in_endpointAddress;
	//	printk(" interrupt endpoint:%d \n",mos78xx_serial->interrupt_in_endpoint);
		mos78xx_serial->interrupt_read_urb = serial->port[0].interrupt_in_urb;
		//printk(" interval:%d\n",mos78xx_serial->interrupt_read_urb->interval);
	  	mos78xx_serial->interrupt_read_urb->transfer_buffer_length = 5;
		//printk(" length:%d\n",mos78xx_serial->interrupt_read_urb->transfer_buffer_length);
		FILL_INT_URB(                                   \
				mos78xx_serial->interrupt_read_urb,     \
				serial->dev,                            \
				usb_rcvintpipe(serial->dev,mos78xx_serial->interrupt_in_endpoint),    \
				mos78xx_serial->interrupt_in_buffer,             \
				mos78xx_serial->interrupt_read_urb->transfer_buffer_length,\
				mos78xx_read_int_callback, mos78xx_serial,     \
				mos78xx_serial->interrupt_read_urb->interval  );

		/* start interrupt read for mos78xx               *
		 * will continue as long as mos78xx is connected  */

		response = usb_submit_urb (mos78xx_serial->interrupt_read_urb);
		if (response)
		{
			printk("%s - Error %d submitting interrupt urb", __FUNCTION__, response);
		}
	}

#endif

	DPRINTK("port number is %d \n",port->number);
	DPRINTK("serial number is %d \n",port->serial->minor);
	DPRINTK("Bulkin endpoint is %d \n",port->bulk_in_endpointAddress);
	DPRINTK("BulkOut endpoint is %d \n",port->bulk_out_endpointAddress);
	DPRINTK("Interrupt endpoint is %d \n",port->interrupt_in_endpointAddress);
	DPRINTK("port's number in the device is %d\n",mos78xx_port->port_num);
	
	mos78xx_port->bulk_in_buffer = port->bulk_in_buffer;
	mos78xx_port->bulk_in_endpoint  = port->bulk_in_endpointAddress;
	mos78xx_port->read_urb          = port->read_urb;
	mos78xx_port->bulk_out_endpoint = port->bulk_out_endpointAddress;
	memset(mos78xx_port->bulk_in_buffer,0,mos78xx_port->read_urb->transfer_buffer_length);

	if((mos78xx_serial->mos78xx_spectrum_2or4ports==2)&&((__u16)port->number != 0))
	{
		FILL_BULK_URB(
				mos78xx_port->read_urb,serial->dev,\
				usb_rcvbulkpipe(serial->dev, (port->bulk_in_endpointAddress+2)),\
				mos78xx_port->bulk_in_buffer,\
				mos78xx_port->read_urb->transfer_buffer_length,         \
				mos78xx_read_bulk_callback,mos78xx_port);
	}
	else
	{
		FILL_BULK_URB(
				mos78xx_port->read_urb,                                 \
				serial->dev,                                            \
				usb_rcvbulkpipe(serial->dev, port->bulk_in_endpointAddress),\
				mos78xx_port->bulk_in_buffer,                                   \
				mos78xx_port->read_urb->transfer_buffer_length,         \
				mos78xx_read_bulk_callback,mos78xx_port);
	}                                                                       
	response = usb_submit_urb (mos78xx_port->read_urb);
	if (response)
	{
		printk("%s - Error %d submitting control urb", __FUNCTION__, response);
	}
	/* initialize our wait queues */
	init_waitqueue_head(&mos78xx_port->wait_open);
	init_waitqueue_head(&mos78xx_port->wait_chase);
	init_waitqueue_head(&mos78xx_port->delta_msr_wait);
	init_waitqueue_head(&mos78xx_port->delta_cts_wait);
	mos78xx_port->delta_cts_cond = 1;
	init_waitqueue_head(&mos78xx_port->xoff_wait);
	mos78xx_port->xoff_cond = 1;
	init_waitqueue_head(&mos78xx_port->wait_command);

	/* initialize our icount structure */
	memset (&(mos78xx_port->icount), 0x00, sizeof(mos78xx_port->icount));

	/* initialize our port settings */
	mos78xx_port->shadowMCR  = MCR_MASTER_IE; /* Must set to enable ints! */        mos78xx_port->chaseResponsePending = FALSE;
	/* send a open port command */
	mos78xx_port->openPending = FALSE;
	mos78xx_port->open        = TRUE;

	if (port->tty) {
		mos78xx_set_termios (port, &tmp_termios);
	}

	mos78xx_port->rxBytesAvail = 0x0;
	mos78xx_port->icount.tx=0;
	mos78xx_port->icount.rx=0;
	return 0;
}

static int mos78xx_chars_in_buffer (struct usb_serial_port *port)
{
	int i;
	int chars = 0;
	struct moschip_port *mos78xx_port;
	dbg("%s - port %d", __FUNCTION__, port->number);
	mos78xx_port = port->private;
	if (mos78xx_port == NULL)
	{
		DPRINTK("%s \n","mos78xx_break:leaving ...........");
		return -1;
	}

	for (i = 0; i < NUM_URBS; ++i)
	{
		if (mos78xx_port->write_urb_pool[i]->status == -EINPROGRESS)
		{
			chars += URB_TRANSFER_BUFFER_SIZE;
		}
	}

	dbg("%s - returns %d", __FUNCTION__, chars);
	return (chars);
}

static void mos78xx_block_until_tx_empty (struct moschip_port *mos78xx_port)
{
	int timeout = HZ/10;
	int wait = 30;
	int count;
	while (1)
	{
		count = mos78xx_chars_in_buffer(mos78xx_port->port);
		if(count<=0)
		{
			return;
		}

		/* Block the thread for a while */
		interruptible_sleep_on_timeout (&mos78xx_port->wait_chase, timeout);
		/* No activity.. count down section */
		wait--;
		if (wait == 0)
		{
			dbg("%s - TIMEOUT", __FUNCTION__);
			return;
		}
		else
		{
			/* Reset timout value back to seconds */
			wait = 30;
		}
	}
}

static void mos78xx_close (struct usb_serial_port *port, struct file * filp)
{
	struct usb_serial *serial;
	struct moschip_serial *mos78xx_serial;
	struct moschip_port *mos78xx_port;
	int status;
	unsigned int no_urbs;
	__u16 Data;
	dbg("%s - port %d", __FUNCTION__, port->number);
//	serial = get_usb_serial (port, __FUNCTION__);
	serial = port->serial;
	if (!serial)
		return;
	mos78xx_port = port->private;
	if (mos78xx_port == NULL)
	{
		printk("%s mos78xx_port returned NULL\n",__FUNCTION__);
		return;
	}	
	mos78xx_serial = serial->private;
	if(mos78xx_serial == NULL)
	{
		printk("%s mos78xx_serial returned NULL\n",__FUNCTION__);
		return;
	}
	if (serial->dev)
	{
		/* flush and block(wait) until tx is empty*/
		mos78xx_block_until_tx_empty(mos78xx_port);	
	}
	for (no_urbs = 0; no_urbs < NUM_URBS;no_urbs++)
	{
		usb_unlink_urb (mos78xx_port->write_urb_pool[no_urbs]);
	}

	for (no_urbs = 0; no_urbs< NUM_URBS; ++no_urbs)
	{
		if (mos78xx_port->write_urb_pool[no_urbs])
		{
			if (mos78xx_port->write_urb_pool[no_urbs]->transfer_buffer)
				kfree(mos78xx_port->write_urb_pool[no_urbs]->transfer_buffer);
			usb_free_urb (mos78xx_port->write_urb_pool[no_urbs]);
		}
	}
	/* While closing port, shutdown all bulk read, write  *
	 * and interrupt read if they exists                  */
	if (serial->dev)
	{
		if (mos78xx_port->write_urb)
		{
			DPRINTK("%s","Shutdown bulk write\n");
			usb_unlink_urb (mos78xx_port->write_urb);
		}
		if (mos78xx_port->read_urb)
		{
			DPRINTK("%s","Shutdown bulk read\n");
			usb_unlink_urb (mos78xx_port->read_urb);
		}
		if((&mos78xx_port->control_urb))
		{
			DPRINTK("%s","Shutdown control read\n");
		}

	}

	mos78xx_serial->NoOfOpenPorts--;
	DPRINTK("NoOfOpenPorts in close%d:in port%d\n",(mos78xx_serial->NoOfOpenPorts,port->number));
	if(mos78xx_serial->NoOfOpenPorts==0)
	{
		//stop the stus polling here
		//printk("disabling the status polling flag to FALSE :\n");
		mos78xx_serial->status_polling_started = FALSE;
		if(mos78xx_serial->interrupt_read_urb)
		{
			DPRINTK("%s","Shutdown interrupt_read_urb\n",__FUNCTION__);
		}
	}

	if (mos78xx_port->write_urb)
	{
		/* if this urb had a transfer buffer already (old tx) free it */                if (mos78xx_port->write_urb->transfer_buffer != NULL)
		{
			kfree(mos78xx_port->write_urb->transfer_buffer);
		}
		usb_free_urb(mos78xx_port->write_urb);
	}
	// clear the MCR & IER
	Data = 0x00;
	mos78xx_set_Uart_Reg(port,MODEM_CONTROL_REGISTER,Data);
	Data = 0x00;
	mos78xx_set_Uart_Reg(port,INTERRUPT_ENABLE_REGISTER,Data);

	//mos78xx_get_Uart_Reg(port,MODEM_CONTROL_REGISTER,&Data1);
	//printk("value of MCR after closing the port is : 0x%x\n",Data1);
	mos78xx_port->open         = FALSE;
	mos78xx_port->closePending = FALSE;
	mos78xx_port->openPending  = FALSE;
}

static void mos78xx_control_callback (struct urb *urb)
{

	unsigned char *data;
	struct moschip_port *mos78xx_port;
	__u8 regval=0x0;

	if(!urb)
	{
		DPRINTK("%s","Invalid Pointer !!!!:\n"); 
		return;
	}
	switch (urb->status)
	{
		case 0:
			/* success */
			break;
		case -ECONNRESET:
		case -ENOENT:
		case -ESHUTDOWN:
			/* this urb is terminated, clean up */
			dbg("%s - urb shutting down with status: %d", __FUNCTION__, urb->status);
			return;
		default:
			dbg("%s - nonzero urb status received: %d", __FUNCTION__, urb->status);
			goto exit;
	}


	mos78xx_port = (struct moschip_port *)urb->context;
	data=urb->transfer_buffer;
	regval=(__u8)data[0];
	if(mos78xx_port->MsrLsr==0)
		mos78xx_handle_newMsr(mos78xx_port,regval);
	else if(mos78xx_port->MsrLsr==1)
		mos78xx_handle_newLsr(mos78xx_port,regval);
exit:
	return;
}

int mos78xx_handle_newMsr(struct moschip_port *port,__u8 newMsr)
{
	struct moschip_port *mos78xx_port;
	struct  async_icount *icount;
	mos78xx_port=port;
	icount = &mos78xx_port->icount;
	if (newMsr & (MOS_MSR_DELTA_CTS | MOS_MSR_DELTA_DSR | MOS_MSR_DELTA_RI | MOS_MSR_DELTA_CD)) {
		icount = &mos78xx_port->icount;

		/* update input line counters */
		if (newMsr & MOS_MSR_DELTA_CTS) {
			icount->cts++;
		}
		if (newMsr & MOS_MSR_DELTA_DSR) {
			icount->dsr++;
		}
		if (newMsr & MOS_MSR_DELTA_CD) {
			icount->dcd++;
		}
		if (newMsr & MOS_MSR_DELTA_RI) {
			icount->rng++;
		}
	}
	if (newMsr & MOS_MSR_CTS) {
		wake_up(&mos78xx_port->delta_cts_wait);
		mos78xx_port->delta_cts_cond=1;
	}
	else
	{
		mos78xx_port->delta_cts_cond=0;
	}
	return 0;
}

int mos78xx_handle_newLsr(struct moschip_port *port,__u8 newLsr)
{
	struct  async_icount *icount;
	dbg("%s - %02x", __FUNCTION__, newLsr);
	if (newLsr & SERIAL_LSR_BI) {
		//
		// Parity and Framing errors only count if they
		// occur exclusive of a break being
		// received.
		//
		newLsr &= (__u8)(SERIAL_LSR_OE | SERIAL_LSR_BI);
	}
	/* update input line counters */
	icount = &port->icount;
	if (newLsr & SERIAL_LSR_BI) {
		icount->brk++;
	}
	if (newLsr & SERIAL_LSR_OE) {
		icount->overrun++;
	}
	if (newLsr & SERIAL_LSR_PE) {
		icount->parity++;
	}
	if (newLsr & SERIAL_LSR_FE) {
		icount->frame++;
	}
	return 0;
}

static int mos78xx_get_reg(struct moschip_port *mcs,__u16 Wval, __u16 reg, __u16 * val)
{
	struct usb_device *dev = mcs->port->serial->dev;        struct usb_ctrlrequest *dr=NULL;
	unsigned char  *buffer=NULL;
	int ret=0;
	buffer= (__u8 *)mcs->ctrl_buf;

	//      dr=(struct usb_ctrlrequest *)(buffer);
	dr=(void *)(buffer + 2);
	dr->bRequestType = MCS_RD_RTYPE;
	dr->bRequest = MCS_RDREQ;
	dr->wValue = cpu_to_le16(Wval);//0;
	dr->wIndex = cpu_to_le16(reg);
	dr->wLength = cpu_to_le16(2);

	FILL_CONTROL_URB(mcs->control_urb,dev,usb_rcvctrlpipe(dev,0),(unsigned char *)dr,buffer,2,mos78xx_control_callback,mcs);
	mcs->control_urb->transfer_buffer_length = 2;
	ret=usb_submit_urb(mcs->control_urb);   
	return ret;
}


static int mos78xx_get_reg_sync(struct usb_serial_port *port, __u16 reg, __u16 * val)
{
	struct usb_device *dev = port->serial->dev;
	int ret=0;
	ret = usb_control_msg(dev, usb_rcvctrlpipe(dev, 0), MCS_RDREQ,
			MCS_RD_RTYPE, 0, reg, val, VENDOR_READ_LENGTH,MOS_WDR_TIMEOUT);
	DPRINTK("mos78xx_get_reg_sync offset is %x, return val %x\n",reg,*val);
	*val = (*val) & 0x00ff;
	return ret;
}

static int mos78xx_set_reg_sync(struct usb_serial_port *port, __u16 reg, __u16  val)
{
	struct usb_device *dev = port->serial->dev;
	val = val & 0x00ff;
	DPRINTK("mos78xx_set_reg_sync offset is %x, value %x\n",reg,val);
	return usb_control_msg(dev, usb_sndctrlpipe(dev, 0), MCS_WRREQ,
			MCS_WR_RTYPE, val, reg, NULL, 0,MOS_WDR_TIMEOUT);
}


static int mos78xx_get_Uart_Reg(struct usb_serial_port *port, __u16 reg, __u16 * val)
{
	struct usb_device *dev = port->serial->dev;
	struct usb_serial *serial;
	int ret=0;
	__u16 Wval;
	struct moschip_serial *mos78xx_serial;
//	serial = get_usb_serial (port, __FUNCTION__);
	serial = port->serial;
	if (!serial)
		return;
	mos78xx_serial = serial->private;

	if(mos78xx_serial->mos78xx_spectrum_2or4ports ==4)
	{
		Wval=(((__u16)port->number - (__u16)(port->serial->minor))+1)<<8;
	}
	else
	{
		if( ((__u16)port->number - (__u16)(port->serial->minor)) == 0) 
		{
			Wval=(((__u16)port->number - (__u16)(port->serial->minor))+1)<<8;
		}
		else
		{
			Wval=(((__u16)port->number - (__u16)(port->serial->minor))+2)<<8;
		}
	}
	ret = usb_control_msg(dev, usb_rcvctrlpipe(dev, 0), MCS_RDREQ,
			MCS_RD_RTYPE, Wval, reg, val,VENDOR_READ_LENGTH,MOS_WDR_TIMEOUT);
	*val = (*val) & 0x00ff;
	return ret;

}


static int mos78xx_set_Uart_Reg(struct usb_serial_port *port, __u16 reg, __u16
val)
{
	struct usb_device *dev = port->serial->dev;
	struct usb_serial *serial;
	struct moschip_serial *mos78xx_serial;
//	serial = get_usb_serial (port, __FUNCTION__);
	serial = port->serial;
	if (!serial)
		return;
	mos78xx_serial = serial->private;

	val = val & 0x00ff;
	// For the UART control registers, the application number need to be Or'ed
	if(mos78xx_serial->mos78xx_spectrum_2or4ports == 4)
	{
		val |= (((__u16)port->number - (__u16)(port->serial->minor))+1)<<8;
	}
	else
	{
		if( ((__u16)port->number - (__u16)(port->serial->minor)) == 0)
		{
			val |= (((__u16)port->number - (__u16)(port->serial->minor))+1)<<8;
		}
		else
		{
			val |= (((__u16)port->number - (__u16)(port->serial->minor))+2)<<8;
		}
	}
	return usb_control_msg(dev, usb_sndctrlpipe(dev, 0), MCS_WRREQ,
			MCS_WR_RTYPE, val, reg, NULL, 0,MOS_WDR_TIMEOUT);
}

static int mos78xx_send_cmd_write_baud_rate (struct moschip_port *mos78xx_port,
int baudRate)
{
	int divisor = 0;
	int status;
	__u16 Data;
	unsigned char number ;
	__u16 clk_sel_val;
	struct usb_serial_port *port;
	if (mos78xx_port == NULL)
		return -1;
	port = (struct usb_serial_port*)mos78xx_port->port;
	dbg("%s - port %d", __FUNCTION__, port->number);
	number = mos78xx_port->port->number - mos78xx_port->port->serial->minor;
	dbg("%s - port = %d, baud = %d", __FUNCTION__, mos78xx_port->port->number, baudRate);
	//reset clk_uart_sel in spregOffset
	if(baudRate >115200)
	{
#ifdef HW_flow_control
		//NOTE: need to see the pther register to modify
		//setting h/w flow control bit to 1;
		status=0;
		//Data = mos78xx_port->shadowMCR ;
		Data = 0x2b;
		mos78xx_port->shadowMCR=Data;
		status=mos78xx_set_Uart_Reg(port,MODEM_CONTROL_REGISTER,Data);
		if(status<0)
		{
			DPRINTK("Writing spreg failed in set_serial_baud\n");
			return -1;
		}
#endif

	}
	else
	{
#ifdef HW_flow_control
		//setting h/w flow control bit to 0;
		status=0;
		//Data = mos78xx_port->shadowMCR ;
		Data = 0xb;
		mos78xx_port->shadowMCR=Data;
		status=mos78xx_set_Uart_Reg(port,MODEM_CONTROL_REGISTER,Data);
		if(status<0)
		{
			DPRINTK("Writing spreg failed in set_serial_baud\n");
			return -1;
		}
#endif
	}
	if(1)//baudRate <= 115200)
	{
		clk_sel_val=0x0;
		Data=0x0;
		status=0;
		status = mos78xx_calc_baud_rate_divisor (baudRate, &divisor,&clk_sel_val);
		status= mos78xx_get_reg_sync(port,mos78xx_port->SpRegOffset,&Data);
		if(status<0)
		{
			DPRINTK("reading spreg failed in set_serial_baud\n");
			return -1;
		}
		Data = (Data & 0x8f)|clk_sel_val;
		status=0;
		status= mos78xx_set_reg_sync(port,mos78xx_port->SpRegOffset,Data);
		if(status<0)
		{
			DPRINTK("Writing spreg failed in set_serial_baud\n");
			return -1;
		}
		/* Calculate the Divisor */
		if (status)
		{
			err("%s - bad baud rate", __FUNCTION__);
			return status;
		}
		/* Enable access to divisor latch */
		Data = mos78xx_port->shadowLCR | SERIAL_LCR_DLAB;
		mos78xx_port->shadowLCR  = Data;
		mos78xx_set_Uart_Reg(port,LINE_CONTROL_REGISTER,Data);
		/* Write the divisor */
		Data = LOW8 (divisor);//MOSCHIP:  commented to test
		DPRINTK("set_serial_baud Value to write DLL is %x\n",Data);
		mos78xx_set_Uart_Reg(port,DIVISOR_LATCH_LSB,Data);
		Data = HIGH8 (divisor); //MOSCHIP:  commented to test
		DPRINTK("set_serial_baud Value to write DLM is %x\n",Data);
		mos78xx_set_Uart_Reg(port,DIVISOR_LATCH_MSB,Data);

		/* Disable access to divisor latch */
		Data = mos78xx_port->shadowLCR & ~SERIAL_LCR_DLAB;
		mos78xx_port->shadowLCR = Data;
		mos78xx_set_Uart_Reg(port,LINE_CONTROL_REGISTER,Data);
	}
	return status;
}

static int mos78xx_calc_baud_rate_divisor (int baudRate, int *divisor,__u16 *clk_sel_val)
{
	dbg("%s - %d", __FUNCTION__, baudRate);

	if(baudRate <=115200)
	{
		*divisor = 115200/baudRate;
		*clk_sel_val = 0x0;
	}
	if((baudRate > 115200) && (baudRate <= 230400))
	{
		*divisor = 230400/baudRate;	
		*clk_sel_val=0x10;
	}
	else if((baudRate > 230400) && (baudRate <= 403200))
	{
		*divisor = 403200/baudRate;	
		*clk_sel_val=0x20;
	}
	else if((baudRate > 403200) && (baudRate <= 460800))
	{
		*divisor = 460800/baudRate;	
		*clk_sel_val=0x30;
	}
	else if((baudRate > 460800) && (baudRate <= 806400))
	{
		*divisor = 806400/baudRate;	
		*clk_sel_val=0x40;
	}
	else if((baudRate >806400) && (baudRate <= 921600))
	{
		*divisor = 921600/baudRate;	
		*clk_sel_val=0x50;
	}
	else if((baudRate > 921600) && (baudRate <= 1572864))
	{
		*divisor = 1572864/baudRate;	
		*clk_sel_val=0x60;
	}
	else if((baudRate > 1572864) && (baudRate <= 3145728))
	{
		*divisor = 3145728/baudRate;	
		*clk_sel_val=0x70;
	}
	return 0;	

#ifdef NOTMCS7840

	for (i = 0; i < NUM_ENTRIES(mos78xx_divisor_table); i++) 
	{
		if ( mos78xx_divisor_table[i].BaudRate == baudrate ) 
		{
			*divisor = mos78xx_divisor_table[i].Divisor;
			return 0;
		}
	}

	/* After trying for all the standard baud rates    *
	 * Try calculating the divisor for this baud rate  */

	if (baudrate > 75 &&  baudrate < 230400) 
	{
		/* get the divisor */
		custom = (__u16)(230400L  / baudrate);

		/* Check for round off */
		round1 = (__u16)(2304000L / baudrate);
		round = (__u16)(round1 - (custom * 10));
		if (round > 4) {
			custom++;
		}
		*divisor = custom;

		DPRINTK(" Baud %d = %d\n",baudrate, custom);
		return 0;
	}

	DPRINTK("%s\n"," Baud calculation Failed...");
	return -1;
#endif
}

static int mos78xx_calc_num_ports(struct usb_serial *serial)
{

	__u16 Data=0x00;
	int ret =0;
	int mos78xx_2or4ports;
	ret = usb_control_msg(serial->dev, usb_rcvctrlpipe(serial->dev, 0),MCS_RDREQ,MCS_RD_RTYPE,0,GPIO_REGISTER,&Data,VENDOR_READ_LENGTH,MOS_WDR_TIMEOUT);
	if((Data&0x01)==0)
	{
		mos78xx_2or4ports =2;
		serial->type->num_bulk_in=2;
		serial->type->num_bulk_out=2;
		serial->type->num_ports=2;
	}
	//else if(serial->interface->cur_altsetting->desc.bNumEndpoints == 9)
	else
	{
		mos78xx_2or4ports =4;
		serial->type->num_bulk_in=4;
		serial->type->num_bulk_out=4;
		serial->type->num_ports=4;
	}
	return mos78xx_2or4ports;
}

static int __init mos78xx_init (void)
{
	usb_serial_register (&mos78xx_device);
	info(DRIVER_DESC " " DRIVER_VERSION);
	return 0;
}
static void __exit mos78xx_exit (void)
{
	usb_serial_deregister (&mos78xx_device);
}

module_init(mos78xx_init);
module_exit(mos78xx_exit);

MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");

